==> dezero-0.0.9.tar.gz: dezero-0.0.9/dezero/functions.py <==
import numpy as np
import dezero
from dezero import cuda, utils
from dezero.core import Function, Variable, as_variable


# =============================================================================
# sin / cos / tanh / exp / log
# =============================================================================
class Sin(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.sin(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy * cos(x)
        return gx


def sin(x):
    f = Sin()
    return f(x)


class Cos(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.cos(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy * -sin(x)
        return gx


def cos(x):
    f = Cos()
    return f(x)


class Tanh(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        return xp.tanh(x)

    def backward(self, gy):
        y = self.outputs[0]()  # weakref
        gx = gy * (1 - y * y)
        return gx


def tanh(x):
    f = Tanh()
    return f(x)


class Exp(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        return xp.exp(x)

    def backward(self, gy):
        y = self.outputs[0]()  # weakref
        gx = gy * y
        return gx


def exp(x):
    f = Exp()
    return f(x)


class Log(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        return xp.log(x)

    def backward(self, gy):
        x, = self.inputs
        gx = gy / x
        return gx


def log(x):
    f = Log()
    return f(x)


# =============================================================================
# for Tensor: sum / repeat / reshape / sum_to / broadcast_to / get_item
# =============================================================================
class Reshape(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        return x.reshape(self.shape)

    def backward(self, gy):
        return reshape(gy, self.x_shape)


def reshape(x, shape):
    x = as_variable(x)
    if x.shape == shape:
        return x
    f = Reshape(shape)
    return f(x)


def expand_dims(x, axis):
    x = as_variable(x)
    shape = list(x.shape)
    shape.insert(axis, 1)
    return reshape(x, tuple(shape))


class Sum(Function):
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        y = x.sum(axis=self.axis, keepdims=self.keepdims)
        return y

    def backward(self, gy):
        x, = self.inputs
        shape = utils.sum_backward_shape(gy, x, self.axis, self.keepdims)
        gy = reshape(gy, shape)
        gx = broadcast_to(gy, x.shape)
        return gx


def sum(x, axis=None, keepdims=False):
    f = Sum(axis, keepdims)
    return f(x)


class SumTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        y = utils.sum_to(x, self.shape)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = broadcast_to(gy, x.shape)
        return gx


def sum_to(x, shape):
    if x.shape == shape:
        return as_variable(x)
    f = SumTo(shape)
    return f(x)


class BroadcastTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        xp = dezero.cuda.get_array_module(x)
        y = xp.broadcast_to(x, self.shape)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = sum_to(gy, x.shape)
        return gx


def broadcast_to(x, shape):
    if x.shape == shape:
        return as_variable(x)
    f = BroadcastTo(shape)
    return f(x)


class MatMul(Function):
    def forward(self, x, W):
        y = x.dot(W)
        return y

    def backward(self, gy):
        x, W = self.inputs
        gx = matmul(gy, transpose(W))
        gW = matmul(transpose(x), gy)
        return gx, gW


def matmul(x, W):
    f = MatMul()
    return f(x, W)


def linear(x, W, b=None):
    x, W = as_variable(x), as_variable(W)
    t = matmul(x, W)
    if b is None:
        return t

    y = t + b
    # t の ndarray はリリースできる
    # （matmul関数の逆伝播では出力データは不要、かつadd関数の逆伝播では入力データは不要）
    t.data = None
    return y


class Transpose(Function):
    def __init__(self, axes=None):
        self.axes = axes

    def forward(self, x):
        y = x.transpose(self.axes)
        return y

    def backward(self, gy):
        if self.axes is None:
            return transpose(gy)

        axes_len = len(self.axes)
        inv_axes = tuple(np.argsort([ax % axes_len for ax in self.axes]))
        return transpose(gy, inv_axes)


def transpose(x, axes=None):
    f = Transpose(axes)
    return f(x)


class GetItem(Function):
    def __init__(self, slices):
        self.slices = slices

    def forward(self, x):
        return x[self.slices]

    def backward(self, gy):
        x, = self.inputs
        return GetItemGrad(self.slices, x.shape)(gy)


class GetItemGrad(Function):
    def __init__(self, slices, in_shape):
        self.slices = slices
        self.in_shape = in_shape

    def forward(self, gy):
        xp = dezero.cuda.get_array_module(gy)
        gx = xp.zeros(self.in_shape, dtype=gy.dtype)

        if xp is np:
            np.add.at(gx, self.slices, gy)
        else:
            xp.scatter_add(gx, self.slices, gy)
        return gx

    def backward(self, ggx):
        return get_item(ggx, self.slices)


def get_item(x, slices):
    f = GetItem(slices)
    return f(x)


# =============================================================================
# activation / loss function
# =============================================================================
def mean_squared_error(y1, y2):
    y1, y2 = as_variable(y1), as_variable(y2)

    N = y1.shape[0]
    diff = y1 - y2
    loss = sum(diff * diff) / N
    return loss


class Sigmoid(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = 1 / (1 + xp.exp(-x))
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = gy * y * (1-y)
        return gx


def sigmoid(x):
    f = Sigmoid()
    return f(x)


def sigmoid_naive(x):
    x = as_variable(x)
    y = 1 / (1 + exp(-x))
    return y


class ReLU(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.maximum(x, 0.0)
        return y

    def backward(self, gy):
        x, = self.inputs
        mask = x.data > 0
        gx = gy * mask
        return gx


def relu(x):
    f = ReLU()
    return f(x)


class Softmax(Function):
    def __init__(self, axis=1):
        self.axis = axis

    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = x - x.max(axis=self.axis, keepdims=True)
        y = xp.exp(y)
        y /= y.sum(axis=self.axis, keepdims=True)
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = y * gy
        sumdx = gx.sum(axis=self.axis, keepdims=True)
        gx -= y * sumdx
        return gx


def softmax(x, axis=1):
    f = Softmax(axis)
    return f(x)


def softmax_naive(x, axis=1):
    x = as_variable(x)
    y = exp(x)
    sum_shape = list(y.shape)
    sum_shape[axis] = 1
    sum_y = sum_to(y, sum_shape)
    return y / sum_y


def softmax_cross_entropy(x, t):
    x, t = as_variable(x), as_variable(t)
    N = x.shape[0]

    p = softmax(x)
    log_p = log(p)
    tlog_p = log_p[np.arange(N), t.data]
    y = -1 * sum(tlog_p) / N
    return y


def accuracy(y, t):
    """
    [WAR] この関数は微分可能ではありません
    """
    y, t = as_variable(y), as_variable(t)

    pred = y.data.argmax(axis=1).reshape(t.shape)
    result = (pred == t.data)
    acc = result.mean()
    return Variable(acc)


# =============================================================================
# embed_id / dropout / batch_norm
# =============================================================================
def embed_id(x, W):
    return W[x]


def dropout(x, dropout_ratio=0.5):
    x = as_variable(x)

    if dezero.config.train:
        xp = cuda.get_array_module(x)
        mask = xp.random.rand(*x.shape) > dropout_ratio
        scale = 1.0 - dropout_ratio
        y = x * mask / scale
        return y
    else:
        return x


def batch_nrom(x):
    pass


# =============================================================================
# max / min
# =============================================================================
class Max(Function):
    def __init__(self, axis=None, keepdims=False):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        return x.max(axis=self.axis, keepdims=self.keepdims)

    def backward(self, gy):
        x = self.inputs[0]
        y = self.outputs[0]()  # weakref

        shape = utils.max_backward_shape(x, self.axis)
        gy = reshape(gy, shape)
        y = reshape(y, shape)
        cond = (x.data == y.data)
        gy = broadcast_to(gy, cond.shape)
        return gy * cond


class Min(Max):
    def forward(self, x):
        return x.min(axis=self.axis, keepdims=self.keepdims)


def max(x, axis=None, keepdims=False):
    f = Max(axis, keepdims)
    return f(x)


def min(x, axis=None, keepdims=False):
    f = Min(axis, keepdims)
    return f(x)


# =============================================================================
#conv2d / col2im / im2col / basic_math（他ファイルの関数をfunciontsへインポート）
# =============================================================================
from dezero.functions_conv import conv2d
from dezero.functions_conv import deconv2d
from dezero.functions_conv import conv2d_simple
from dezero.functions_conv import im2col
from dezero.functions_conv import col2im
from dezero.core import add
from dezero.core import sub
from dezero.core import rsub
from dezero.core import mul
from dezero.core import div
from dezero.core import neg
from dezero.core import pow

==> dezero-0.0.10.tar.gz: dezero-0.0.10/dezero/functions.py <==
import numpy as np
import dezero
from dezero import cuda, utils
from dezero.core import Function, Variable, as_variable


# =============================================================================
# sin / cos / tanh / exp / log
# =============================================================================
class Sin(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.sin(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy * cos(x)
        return gx


def sin(x):
    f = Sin()
    return f(x)


class Cos(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.cos(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy * -sin(x)
        return gx


def cos(x):
    f = Cos()
    return f(x)


class Tanh(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        return xp.tanh(x)

    def backward(self, gy):
        y = self.outputs[0]()  # weakref
        gx = gy * (1 - y * y)
        return gx


def tanh(x):
    f = Tanh()
    return f(x)


class Exp(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        return xp.exp(x)

    def backward(self, gy):
        y = self.outputs[0]()  # weakref
        gx = gy * y
        return gx


def exp(x):
    f = Exp()
    return f(x)


class Log(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        return xp.log(x)

    def backward(self, gy):
        x, = self.inputs
        gx = gy / x
        return gx


def log(x):
    f = Log()
    return f(x)


# =============================================================================
# Tensor operations: sum / repeat / reshape / sum_to / broadcast_to / get_item
# =============================================================================
class Reshape(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        return x.reshape(self.shape)

    def backward(self, gy):
        return reshape(gy, self.x_shape)


def reshape(x, shape):
    x = as_variable(x)
    if x.shape == shape:
        return x
    f = Reshape(shape)
    return f(x)


def expand_dims(x, axis):
    x = as_variable(x)
    shape = list(x.shape)
    shape.insert(axis, 1)
    return reshape(x, tuple(shape))


class Sum(Function):
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        y = x.sum(axis=self.axis, keepdims=self.keepdims)
        return y

    def backward(self, gy):
        x, = self.inputs
        shape = utils.sum_backward_shape(gy, x, self.axis, self.keepdims)
        gy = reshape(gy, shape)
        gx = broadcast_to(gy, x.shape)
        return gx


def sum(x, axis=None, keepdims=False):
    f = Sum(axis, keepdims)
    return f(x)


class SumTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        y = utils.sum_to(x, self.shape)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = broadcast_to(gy, x.shape)
        return gx


def sum_to(x, shape):
    if x.shape == shape:
        return as_variable(x)
    f = SumTo(shape)
    return f(x)


class BroadcastTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        xp = dezero.cuda.get_array_module(x)
        y = xp.broadcast_to(x, self.shape)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = sum_to(gy, x.shape)
        return gx


def broadcast_to(x, shape):
    if x.shape == shape:
        return as_variable(x)
    f = BroadcastTo(shape)
    return f(x)


class MatMul(Function):
    def forward(self, x, W):
        y = x.dot(W)
        return y

    def backward(self, gy):
        x, W = self.inputs
        gx = matmul(gy, transpose(W))
        gW = matmul(transpose(x), gy)
        return gx, gW


def matmul(x, W):
    f = MatMul()
    return f(x, W)


def linear(x, W, b=None):
    x, W = as_variable(x), as_variable(W)
    t = matmul(x, W)
    if b is None:
        return t

    y = t + b
    # t の ndarray はリリースできる
    # （matmul関数の逆伝播では出力データは不要、かつadd関数の逆伝播では入力データは不要）
    t.data = None
    return y


class Transpose(Function):
    def __init__(self, axes=None):
        self.axes = axes

    def forward(self, x):
        y = x.transpose(self.axes)
        return y

    def backward(self, gy):
        if self.axes is None:
            return transpose(gy)

        axes_len = len(self.axes)
        inv_axes = tuple(np.argsort([ax % axes_len for ax in self.axes]))
        return transpose(gy, inv_axes)


def transpose(x, axes=None):
    f = Transpose(axes)
    return f(x)


class GetItem(Function):
    def __init__(self, slices):
        self.slices = slices

    def forward(self, x):
        return x[self.slices]

    def backward(self, gy):
        x, = self.inputs
        return GetItemGrad(self.slices, x.shape)(gy)


class GetItemGrad(Function):
    def __init__(self, slices, in_shape):
        self.slices = slices
        self.in_shape = in_shape

    def forward(self, gy):
        xp = dezero.cuda.get_array_module(gy)
        gx = xp.zeros(self.in_shape, dtype=gy.dtype)

        if xp is np:
            np.add.at(gx, self.slices, gy)
        else:
            xp.scatter_add(gx, self.slices, gy)
        return gx

    def backward(self, ggx):
        return get_item(ggx, self.slices)


def get_item(x, slices):
    f = GetItem(slices)
    return f(x)


# =============================================================================
# activation / loss function
# =============================================================================
def mean_squared_error(y1, y2):
    y1, y2 = as_variable(y1), as_variable(y2)

    N = y1.shape[0]
    diff = y1 - y2
    loss = sum(diff * diff) / N
    return loss


class Sigmoid(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = 1 / (1 + xp.exp(-x))
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = gy * y * (1-y)
        return gx


def sigmoid(x):
    f = Sigmoid()
    return f(x)


def sigmoid_naive(x):
    x = as_variable(x)
    y = 1 / (1 + exp(-x))
    return y


class ReLU(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.maximum(x, 0.0)
        return y

    def backward(self, gy):
        x, = self.inputs
        mask = x.data > 0
        gx = gy * mask
        return gx


def relu(x):
    f = ReLU()
    return f(x)


class Softmax(Function):
    def __init__(self, axis=1):
        self.axis = axis

    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = x - x.max(axis=self.axis, keepdims=True)
        y = xp.exp(y)
        y /= y.sum(axis=self.axis, keepdims=True)
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = y * gy
        sumdx = gx.sum(axis=self.axis, keepdims=True)
        gx -= y * sumdx
        return gx


def softmax(x, axis=1):
    f = Softmax(axis)
    return f(x)


def softmax_naive(x, axis=1):
    x = as_variable(x)
    y = exp(x)
    sum_shape = list(y.shape)
    sum_shape[axis] = 1
    sum_y = sum_to(y, sum_shape)
    return y / sum_y


def softmax_cross_entropy(x, t):
    x, t = as_variable(x), as_variable(t)
    N = x.shape[0]

    p = softmax(x)
    log_p = log(p)
    tlog_p = log_p[np.arange(N), t.data]
    y = -1 * sum(tlog_p) / N
    return y


def accuracy(y, t):
    """
    [WAR] この関数は微分可能ではありません
    """
    y, t = as_variable(y), as_variable(t)

    pred = y.data.argmax(axis=1).reshape(t.shape)
    result = (pred == t.data)
    acc = result.mean()
    return Variable(acc)


# =============================================================================
# embed_id / dropout / batch_norm
# =============================================================================
def embed_id(x, W):
    return W[x]


def dropout(x, dropout_ratio=0.5):
    x = as_variable(x)

    if dezero.config.train:
        xp = cuda.get_array_module(x)
        mask = xp.random.rand(*x.shape) > dropout_ratio
        scale = 1.0 - dropout_ratio
        y = x * mask / scale
        return y
    else:
        return x


def batch_nrom(x):
    pass


# =============================================================================
# max / min
# =============================================================================
class Max(Function):
    def __init__(self, axis=None, keepdims=False):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        return x.max(axis=self.axis, keepdims=self.keepdims)

    def backward(self, gy):
        x = self.inputs[0]
        y = self.outputs[0]()  # weakref

        shape = utils.max_backward_shape(x, self.axis)
        gy = reshape(gy, shape)
        y = reshape(y, shape)
        cond = (x.data == y.data)
        gy = broadcast_to(gy, cond.shape)
        return gy * cond


class Min(Max):
    def forward(self, x):
        return x.min(axis=self.axis, keepdims=self.keepdims)


def max(x, axis=None, keepdims=False):
    f = Max(axis, keepdims)
    return f(x)


def min(x, axis=None, keepdims=False):
    f = Min(axis, keepdims)
    return f(x)


# =============================================================================
#conv2d / col2im / im2col / basic_math（他ファイルの関数をfunciontsへインポート）
# =============================================================================
from dezero.functions_conv import conv2d
from dezero.functions_conv import deconv2d
from dezero.functions_conv import conv2d_simple
from dezero.functions_conv import im2col
from dezero.functions_conv import col2im
from dezero.functions_conv import pooling_simple
from dezero.functions_conv import pooling
from dezero.core import add
from dezero.core import sub
from dezero.core import rsub
from dezero.core import mul
from dezero.core import div
from dezero.core import neg
from dezero.core import pow
